@@ -9,7 +9,7 @@
 #
 #  The original author is Jochen Wiedmann.
 #
-#  Copyright (C) 2009 by H.Merijn Brand & Jens Rehsack
+#  Copyright (C) 2009,2010 by H.Merijn Brand & Jens Rehsack
 #  Copyright (C) 2004 by Jeff Zucker
 #  Copyright (C) 1998 by Jochen Wiedmann
 #
@@ -31,9 +31,10 @@ package DBD::File;
 
 use strict;
 
+use Carp;
 use vars qw( @ISA $VERSION $drh $valid_attrs );
 
-$VERSION = "0.36";
+$VERSION = "0.38";
 
 $drh = undef;		# holds driver handle(s) once initialised
 
@@ -54,7 +55,8 @@ sub driver ($;$)
     $attr ||= {};
     {	no strict "refs";
 	unless ($attr->{Attribution}) {
-	    $class eq "DBD::File" and $attr->{Attribution} = "$class by Jeff Zucker";
+	    $class eq "DBD::File" and
+		$attr->{Attribution} = "$class by Jeff Zucker";
 	    $attr->{Attribution} ||= ${$class . "::ATTRIBUTION"} ||
 		"oops the author of $class forgot to define this";
 	    }
@@ -74,14 +76,11 @@ sub CLONE
 
 sub file2table
 {
-    my ($data, $dir, $file, $file_is_tab) = @_;
+    my ($data, $dir, $file, $file_is_tab, $quoted) = @_;
 
     $file eq "." || $file eq ".."	and return;
 
-    # Fully Qualified File Name
-    my $fqfn = File::Spec->catfile ($dir, $file);
-
-    my ($ext, $req) = ("", 0, 0);
+    my ($ext, $req) = ("", 0);
     if ($data->{f_ext}) {
 	($ext, my $opt) = split m/\//, $data->{f_ext};
 	if ($ext && $opt) {
@@ -90,8 +89,17 @@ sub file2table
 	}
 
     (my $tbl = $file) =~ s/$ext$//i;
+    $file_is_tab and $file = "$tbl$ext";
 
-    $file_is_tab && $file !~ m/$ext$/i and $fqfn .= $ext;
+    # Fully Qualified File Name
+    my $fqfn;
+    unless ($quoted) { # table names are case insensitive in SQL
+	opendir my $dh, $dir or croak "Can't open '$dir': $!";
+	my @f = grep { lc $_ eq lc $file } readdir $dh;
+	@f == 1 and $file = $f[0];
+	closedir $dh or croak "Can't close '$dir': $!";
+	}
+    $fqfn = File::Spec->catfile ($dir, $file);
 
     $file = $fqfn;
     if ($ext) {
@@ -149,10 +157,13 @@ sub connect ($$;$$$)
 		}
 	    }
         $this->{f_valid_attrs} = {
-	    f_version	=> 1,  # DBD::File version
-	    f_dir	=> 1,  # base directory
-	    f_ext	=> "", # file extension
-	    f_tables	=> 1,  # base directory
+	    f_version	=> 1, # DBD::File version
+	    f_dir	=> 1, # base directory
+	    f_ext	=> 1, # file extension
+	    f_schema	=> 1, # schema name
+	    f_tables	=> 1, # base directory
+	    f_lock	=> 1, # Table locking mode
+	    f_encoding	=> 1, # Encoding of the file
 	    };
         $this->{sql_valid_attrs} = {
 	    sql_handler           => 1, # Nano or S:S
@@ -160,7 +171,7 @@ sub connect ($$;$$$)
 	    sql_statement_version => 1, # S:S version
 	    };
 	}
-    $this->STORE ("Active", 1);
+    $this->STORE (Active => 1);
     return set_versions ($this);
     } # connect
 
@@ -168,8 +179,9 @@ sub set_versions
 {
     my $this = shift;
     $this->{f_version} = $DBD::File::VERSION;
-    for (qw( nano_version statement_version)) {
-	$this->{"sql_$_"} = $DBI::SQL::Nano::versions->{$_} || "";
+    for (qw( nano_version statement_version )) {
+	# strip development release version part
+	($this->{"sql_$_"} = $DBI::SQL::Nano::versions->{$_} || "") =~ s/_[0-9]+$//;
 	}
     $this->{sql_handler} = $this->{sql_statement_version}
 	? "SQL::Statement"
@@ -208,7 +220,7 @@ sub data_sources ($;$)
 	$file ne File::Spec->updir () && -d $d and
 	    push @dsns, "DBI:$driver:f_dir=$d";
 	}
-    @dsns;
+    return @dsns;
     } # data_sources
 
 sub disconnect_all
@@ -231,7 +243,7 @@ $DBD::File::db::imp_data_size = 0;
 
 sub ping
 {
-    return (shift->FETCH ("Active")) ? 1 : 0;
+    ($_[0]->FETCH ("Active")) ? 1 : 0;
     } # ping
 
 sub prepare ($$;@)
@@ -274,7 +286,7 @@ sub prepare ($$;@)
 	    $sth->STORE ("NUM_OF_PARAMS", scalar ($stmt->params ()));
 	    }
 	}
-    $sth;
+    return $sth;
     } # prepare
 
 sub csv_cache_sql_parser_object
@@ -294,8 +306,8 @@ sub csv_cache_sql_parser_object
 
 sub disconnect ($)
 {
-    shift->STORE ("Active", 0);
-    1;
+    $_[0]->STORE (Active => 0);
+    return 1;
     } # disconnect
 
 sub FETCH ($$)
@@ -328,7 +340,7 @@ sub STORE ($$$)
     if ($attrib eq lc $attrib) {
 	# Driver private attributes are lower cased
 
-	# I'm not implementing this yet becuase other drivers may be
+	# I'm not implementing this yet because other drivers may be
 	# setting f_ and sql_ attrs I don't know about
 	# I'll investigate and publicize warnings to DBD authors
 	# then implement this
@@ -337,17 +349,17 @@ sub STORE ($$$)
 	# not implemented yet
 	# my $class = $dbh->FETCH ("ImplementorClass");
 	#
-	# !$dbh->{f_valid_attrs}->{$attrib} && !$dbh->{sql_valid_attrs}->{$attrib} and
+	# !$dbh->{f_valid_attrs}{$attrib} && !$dbh->{sql_valid_attrs}{$attrib} and
 	#    return $dbh->set_err ($DBI::stderr, "Invalid attribute '$attrib'");
 	#  $dbh->{$attrib} = $value;
 
 	if ($attrib eq "f_dir") {
 	    -d $value or
-		return $dbh->set_err ($DBI::stderr, "No such directory '$value'")
+		return $dbh->set_err ($DBI::stderr, "No such directory '$value'");
 	    }
 	if ($attrib eq "f_ext") {
-	    $value eq "" || $value =~ m{^\.\w+(?:/[iIrR]*)?$}
-		or carp "'$value' doesn't look like a valid file extension attribute\n";
+	    $value eq "" || $value =~ m{^\.\w+(?:/[rR]*)?$} or
+		carp "'$value' doesn't look like a valid file extension attribute\n";
 	    }
 	$dbh->{$attrib} = $value;
 	return 1;
@@ -359,6 +371,7 @@ sub DESTROY ($)
 {
     my $dbh = shift;
     $dbh->SUPER::FETCH ("Active") and $dbh->disconnect ;
+    undef $dbh->{csv_sql_parser_object};
     } # DESTROY
 
 sub type_info_all ($)
@@ -417,10 +430,13 @@ sub type_info_all ($)
 	    }
 
 	my ($file, @tables, %names);
-	my $user = eval { getpwuid ((stat _)[4]) };
+	my $schema = exists $dbh->{f_schema}
+	    ? defined $dbh->{f_schema} && $dbh->{f_schema} ne ""
+		? $dbh->{f_schema} : undef
+	    : eval { getpwuid ((stat $dir)[4]) };
 	while (defined ($file = readdir ($dirh))) {
-	    my $tbl = DBD::File::file2table ($dbh, $dir, $file, 0) or next;
-	    push @tables, [ undef, $user, $tbl, "TABLE", undef ];
+	    my $tbl = DBD::File::file2table ($dbh, $dir, $file, 0, 0) or next;
+	    push @tables, [ undef, $schema, $tbl, "TABLE", undef ];
 	    }
 	unless (closedir $dirh) {
 	    $dbh->set_err ($DBI::stderr, "Cannot close directory $dir: $!");
@@ -444,7 +460,7 @@ sub type_info_all ($)
 				    NAMES => $names,
 				    });
 	$sth or $dbh->set_err ($DBI::stderr, $dbh2->errstr);
-	$sth;
+	return $sth;
 	} # table_info
     }
 
@@ -456,7 +472,7 @@ sub list_tables ($)
     while (my $ref = $sth->fetchrow_arrayref ()) {
 	push @tables, $ref->[2];
 	}
-    @tables;
+    return @tables;
     } # list_tables
 
 sub quote ($$;$)
@@ -479,7 +495,7 @@ sub quote ($$;$)
     $str =~ s/\'/\\\'/sg;
     $str =~ s/\n/\\n/sg;
     $str =~ s/\r/\\r/sg;
-    "'$str'";
+    return "'$str'";
     } # quote
 
 sub commit ($)
@@ -487,7 +503,7 @@ sub commit ($)
     my $dbh = shift;
     $dbh->FETCH ("Warn") and
 	carp "Commit ineffective while AutoCommit is on", -1;
-    1;
+    return 1;
     } # commit
 
 sub rollback ($)
@@ -495,7 +511,7 @@ sub rollback ($)
     my $dbh = shift;
     $dbh->FETCH ("Warn") and
 	carp "Rollback ineffective while AutoCommit is on", -1;
-    0;
+    return 0;
     } # rollback
 
 # ====== STATEMENT =============================================================
@@ -511,19 +527,19 @@ sub bind_param ($$$;$)
     my ($sth, $pNum, $val, $attr) = @_;
     if ($attr && defined $val) {
 	my $type = ref $attr eq "HASH" ? $attr->{TYPE} : $attr;
-	if (   $attr == DBI::SQL_BIGINT ()
-	    || $attr == DBI::SQL_INTEGER ()
+	if (   $attr == DBI::SQL_BIGINT   ()
+	    || $attr == DBI::SQL_INTEGER  ()
 	    || $attr == DBI::SQL_SMALLINT ()
-	    || $attr == DBI::SQL_TINYINT ()
-	    ) {
+	    || $attr == DBI::SQL_TINYINT  ()
+	       ) {
 	    $val += 0;
 	    }
 	elsif ($attr == DBI::SQL_DECIMAL ()
-	    || $attr == DBI::SQL_DOUBLE ()
-	    || $attr == DBI::SQL_FLOAT ()
+	    || $attr == DBI::SQL_DOUBLE  ()
+	    || $attr == DBI::SQL_FLOAT   ()
 	    || $attr == DBI::SQL_NUMERIC ()
-	    || $attr == DBI::SQL_REAL ()
-	    ) {
+	    || $attr == DBI::SQL_REAL    ()
+	       ) {
 	    $val += 0.;
 	    }
 	else {
@@ -531,7 +547,7 @@ sub bind_param ($$$;$)
 	    }
 	}
     $sth->{f_params}[$pNum - 1] = $val;
-    1;
+    return 1;
     } # bind_param
 
 sub execute
@@ -541,18 +557,29 @@ sub execute
 
     $sth->finish;
     my $stmt = $sth->{f_stmt};
-    unless ((my $req_prm = scalar($stmt->params())) == (my $nparm = @$params)) {
-	$sth->set_err ($DBI::stderr,
-	    "You passed $nparm parameters where $req_prm required");
-	return;
+    unless ($sth->{f_params_checked}++) {
+	# bug in SQL::Statement 1.20 and below causes breakage
+	# on all but the first call
+	unless ((my $req_prm = $stmt->params ()) == (my $nparm = @$params)) {
+	    my $msg = "You passed $nparm parameters where $req_prm required";
+	    $sth->set_err ($DBI::stderr, $msg);
+	    return;
+	    }
+	}
+    my @err;
+    my $result = eval {
+	local $SIG{__WARN__} = sub { push @err, @_ };
+	$stmt->execute ($sth, $params);
+	};
+    if ($@ || @err) {
+	$sth->set_err ($DBI::stderr, $@ || $err[0]);
+	return undef;
 	}
-    my $result = eval { $stmt->execute ($sth, $params); };
-    $@ and return $sth->set_err ($DBI::stderr, $@);
 
     if ($stmt->{NUM_OF_FIELDS}) {    # is a SELECT statement
 	$sth->STORE (Active => 1);
 	$sth->FETCH ("NUM_OF_FIELDS") or
-	    $sth->STORE ("NUM_OF_FIELDS", $stmt->{NUM_OF_FIELDS})
+	    $sth->STORE ("NUM_OF_FIELDS", $stmt->{NUM_OF_FIELDS});
 	}
     return $result;
     } # execute
@@ -561,29 +588,29 @@ sub finish
 {
     my $sth = shift;
     $sth->SUPER::STORE (Active => 0);
-    delete $sth->{f_stmt}->{data};
+    delete $sth->{f_stmt}{data};
     return 1;
     } # finish
 
 sub fetch ($)
 {
     my $sth  = shift;
-    my $data = $sth->{f_stmt}->{data};
+    my $data = $sth->{f_stmt}{data};
     if (!$data || ref $data ne "ARRAY") {
 	$sth->set_err ($DBI::stderr,
 	    "Attempt to fetch row without a preceeding execute () call or from a non-SELECT statement"
 	    );
-	return
+	return;
 	}
     my $dav = shift @$data;
     unless ($dav) {
 	$sth->finish;
-	return
+	return;
 	}
     if ($sth->FETCH ("ChopBlanks")) {
 	$_ && $_ =~ s/\s+$// for @$dav;
 	}
-    $sth->_set_fbav ($dav);
+    return $sth->_set_fbav ($dav);
     } # fetch
 *fetchrow_arrayref = \&fetch;
 
@@ -626,11 +653,13 @@ sub DESTROY ($)
 {
     my $sth = shift;
     $sth->SUPER::FETCH ("Active") and $sth->finish;
+    undef $sth->{f_stmt};
+    undef $sth->{f_params};
     } # DESTROY
 
 sub rows ($)
 {
-    shift->{f_stmt}->{NUM_OF_ROWS};
+    return $_[0]->{f_stmt}{NUM_OF_ROWS};
     } # rows
 
 package DBD::File::Statement;
@@ -658,7 +687,8 @@ my $open_table_re = sprintf "(?:%s|%s|%s)",
 sub get_file_name ($$$)
 {
     my ($self, $data, $table) = @_;
-    $table =~ s/^\"//;    # handle quoted identifiers
+    my $quoted = 0;
+    $table =~ s/^\"// and $quoted = 1;    # handle quoted identifiers
     $table =~ s/\"$//;
     my $file = $table;
     if (    $file !~ m/^$open_table_re/o
@@ -666,7 +696,8 @@ sub get_file_name ($$$)
 	and $file !~ m{^[a-z]\:}    # drive letter
 	) {
 	exists $data->{Database}{f_map}{$table} or
-	    DBD::File::file2table ($data->{Database}, $data->{Database}{f_dir}, $file, 1);
+	    DBD::File::file2table ($data->{Database},
+		$data->{Database}{f_dir}, $file, 1, $quoted);
 	$file = $data->{Database}{f_map}{$table} || undef;
 	}
     return ($table, $file);
@@ -694,17 +725,28 @@ sub open_table ($$$$$)
 	    $safe_drop or croak "Cannot open $file: $!";
 	    }
 	}
-    $fh and binmode $fh;
-    if ($locking and $fh) {
-	if ($lockMode) {
-	    flock $fh, 2 or
-		croak "Cannot obtain exclusive lock on $file: $!";
+    if ($fh) {
+	if (my $enc = $data->{Database}{f_encoding}) {
+	    binmode $fh, ":encoding($enc)" or
+                croak "Failed to set encoding layer '$enc' on $file: $!";
 	    }
 	else {
-	    flock $fh, 1 or
-		croak "Cannot obtain shared lock on $file: $!";
+	    binmode $fh or croak "Failed to set binary mode on $file: $!";
 	    }
 	}
+    if ($locking and $fh) {
+	my $lm = defined $data->{Database}{f_lock}
+		      && $data->{Database}{f_lock} =~ m/^[012]$/
+		       ? $data->{Database}{f_lock}
+		       : $lockMode ? 2 : 1;
+	if ($lm == 2) {
+	    flock $fh, 2 or croak "Cannot obtain exclusive lock on $file: $!";
+	    }
+	elsif ($lm == 1) {
+	    flock $fh, 1 or croak "Cannot obtain shared lock on $file: $!";
+	    }
+	# $lm = 0 is forced no locking at all
+	}
     my $columns = {};
     my $array   = [];
     my $pos     = $fh ? $fh->tell () : undef;
@@ -718,7 +760,7 @@ sub open_table ($$$$$)
     my $class = ref $self;
     $class =~ s/::Statement/::Table/;
     bless $tbl, $class;
-    $tbl;
+    return $tbl;
     } # open_table
 
 package DBD::File::Table;
@@ -757,7 +799,7 @@ sub truncate ($$)
     my ($self, $data) = @_;
     $self->{fh}->truncate ($self->{fh}->tell ()) or
 	croak "Error while truncating " . $self->{file} . ": $!";
-    1;
+    return 1;
     } # truncate
 
 1;
@@ -787,7 +829,6 @@ See L<DBI> for details on DBI, L<SQL::Statement> for details on
 SQL::Statement and L<DBD::CSV> or L<DBD::IniFile> for example
 drivers.
 
-
 =head2 Metadata
 
 The following attributes are handled by DBI itself and not by DBD::File,
@@ -817,20 +858,20 @@ Works
 
 =item NUM_OF_FIELDS
 
-Valid after C<$sth->execute>
+Valid after C<< $sth->execute >>
 
 =item NUM_OF_PARAMS
 
-Valid after C<$sth->prepare>
+Valid after C<< $sth->prepare >>
 
 =item NAME
 
-Valid after C<$sth->execute>; undef for Non-Select statements.
+Valid after C<< $sth->execute >>; undef for Non-Select statements.
 
 =item NULLABLE
 
 Not really working, always returns an array ref of one's, as DBD::CSV
-doesn't verify input data. Valid after C<$sth->execute>; undef for
+doesn't verify input data. Valid after C<< $sth->execute >>; undef for
 Non-Select statements.
 
 =back
@@ -871,6 +912,67 @@ extension is always case-insensitive. The table names are not.
 In this case the extension is required, and all filenames that do not match
 are ignored.
 
+=item f_schema
+
+This will set the schema name. Default is the owner of the folder in which
+the table file resides.  C<undef> is allowed.
+
+    my $dbh = DBI->connect ("dbi:CSV:", "", "", {
+        f_schema => undef,
+        f_dir    => "data",
+        f_ext    => ".csv/r",
+        }) or die $DBI::errstr;
+
+The effect is that when you get table names from DBI, you can force all
+tables into the same (or no) schema:
+
+    my @tables $dbh->tables ();
+
+    # no f_schema
+    "merijn".foo
+    "merijn".bar
+
+    # f_schema => "dbi"
+    "dbi".foo
+    "dbi".bar
+
+    # f_schema => undef
+    foo
+    bar
+
+Defining f_schema to the empty string is equal to setting it to C<undef>,
+this to enable the DSN to be C<dbi:CSV:f_schema=;f_dir=.>.
+
+=item f_lock
+
+With this attribute, you can force locking mode (if locking is supported
+at all) for opening tables. By default, tables are opened with a shared
+lock for reading, and with an exclusive lock for writing. The supported
+modes are:
+
+=over 2
+
+=item 0
+
+Force no locking at all.
+
+=item 1
+
+Only shared locks will be used.
+
+=item 2
+
+Only exclusive locks will be used.
+
+=back
+
+But see L</"NOWN BUGS"> below.
+
+=item f_encoding
+
+With this attribute, you can set the encoding in which the file is opened.
+This is implemented using C<binmode $fh, ":encoding(<f_encoding>)">.
+
 =back
 
 =head2 Driver private methods
@@ -913,7 +1015,7 @@ MacOS and Windows 95, as there's a single user anyways).
 
 =back
 
-=head1 AUTHOR 
+=head1 AUTHOR
 
 This module is currently maintained by
 
